# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides support for basic commands of G90 alarm panel.
"""
from __future__ import annotations
import logging
import json
import asyncio
from asyncio import Future
from asyncio.protocols import DatagramProtocol
from asyncio.transports import DatagramTransport, BaseTransport
from typing import Optional, Tuple, List, Any, TypeVar, Generic
from dataclasses import dataclass
# `Self` is available in `typing` module only starting from Python 3.11, for
# older versions need to use typing_extensions`
from typing_extensions import Self
from ..exceptions import (
    G90Error, G90TimeoutError, G90CommandFailure, G90CommandError
)
from ..const import G90Commands, G90CommandsBase


_LOGGER = logging.getLogger(__name__)

CommandT = TypeVar('CommandT', bound=G90CommandsBase)
CommandDataT = TypeVar('CommandDataT')


class G90Command(DatagramProtocol, Generic[CommandT, CommandDataT]):
    """
    Base class for command handling for alarm panel protocol.
    """
    # pylint: disable=too-many-instance-attributes
    # Lock need to be shared across all of the class instances
    _sk_lock = asyncio.Lock()

    # pylint: disable=too-many-positional-arguments,too-many-arguments
    def __init__(self, host: str, port: int, code: CommandT,
                 data: Optional[CommandDataT] = None,
                 local_port: Optional[int] = None,
                 timeout: float = 3.0, retries: int = 3) -> None:
        self._remote_host = host
        self._remote_port = port
        self._local_port = local_port
        self._code = code
        self._timeout = timeout
        self._retries = retries
        self._result: Optional[CommandDataT] = None
        self._connection_result: Optional[
            Future[Tuple[str, int, bytes]]
        ] = None
        self._data = self.encode_data(data)

    # Implementation of datagram protocol,
    # https://docs.python.org/3/library/asyncio-protocol.html#datagram-protocols
    def connection_made(self, transport: BaseTransport) -> None:
        """
        Invoked when connection is established.
        """

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """
        Invoked when connection is lost.
        """

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        """
        Invoked when datagram is received.
        """
        if asyncio.isfuture(self._connection_result):
            if self._connection_result.done():
                _LOGGER.debug(
                    'Excessive packet received from %s:%s: %s',
                    addr[0], addr[1], data
                )
                return
            self._connection_result.set_result((*addr, data))

    def error_received(self, exc: Exception) -> None:
        """
        Invoked when error is received.
        """
        if (
            asyncio.isfuture(self._connection_result) and not
            self._connection_result.done()
        ):
            self._connection_result.set_exception(exc)

    async def _create_connection(self) -> (
        Tuple[DatagramTransport, DatagramProtocol]
    ):
        """
        Creates UDP connection to the alarm panel.
        """
        loop = asyncio.get_running_loop()

        _LOGGER.debug('Creating UDP endpoint for %s:%s',
                      self.host, self.port)
        local_addr = None
        if self._local_port:
            local_addr = ('0.0.0.0', self._local_port)

        transport, protocol = await loop.create_datagram_endpoint(
            lambda: self,
            remote_addr=(self.host, self.port),
            allow_broadcast=True,
            local_addr=local_addr)

        return (transport, protocol)

    def encode_data(self, data: Optional[CommandDataT]) -> str:
        """
        Encodes the command data to JSON string.
        """
        raise NotImplementedError()

    def decode_data(self, payload: Optional[str]) -> CommandDataT:
        """
        Decodes the command data from JSON string.
        """
        raise NotImplementedError()

    def to_wire(self) -> bytes:
        """
        Serializes the command to wire format.
        """
        raise NotImplementedError()

    def from_wire(self, data: bytes) -> Optional[CommandDataT]:
        """
        Deserializes the command from wire format.
        """
        raise NotImplementedError()

    @property
    def result(self) -> CommandDataT:
        """
        The result of the command.
        """
        raise NotImplementedError()

    @property
    def host(self) -> str:
        """
        The hostname/IP address of the alarm panel.
        """
        return self._remote_host

    @property
    def port(self) -> int:
        """
        The port of the alarm panel.
        """
        return self._remote_port

    @property
    def expects_response(self) -> bool:
        """
        Indicates whether the command expects a response.
        """
        return True

    async def process(self) -> Self:  # G90Command[CommandT, CommandDataT]:
        """
        Processes the command.
        """
        # Disallow using `NONE` command, which is intended to use by inheriting
        # classes overriding `process()` method
        if self._code == G90Commands.NONE:
            raise G90Error("'NONE' command code is disallowed")

        transport, _ = await self._create_connection()
        attempts = self._retries
        while True:
            attempts = attempts - 1
            loop = asyncio.get_running_loop()
            self._connection_result = loop.create_future()
            async with self._sk_lock:
                _LOGGER.debug('(code %s) Sending request to %s:%s',
                              self._code, self.host, self.port)
                transport.sendto(self.to_wire())
                if not self.expects_response:
                    self._result = None
                    return self
                done, _ = await asyncio.wait([self._connection_result],
                                             timeout=self._timeout)
            if self._connection_result in done:
                break
            # Cancel the future to signal protocol handler it is no longer
            # valid, the future will be re-created on next retry
            self._connection_result.cancel()
            if not attempts:
                transport.close()
                raise G90TimeoutError()
            _LOGGER.debug('Timed out, retrying')
        transport.close()
        (host, port, data) = self._connection_result.result()
        _LOGGER.debug('Received response from %s:%s', host, port)
        if self.host != '255.255.255.255':
            if self.host != host or host == '255.255.255.255':
                raise G90Error(f'Received response from wrong host {host},'
                               f' expected from {self.host}')
        if self.port != port:
            raise G90Error(f'Received response from wrong port {port},'
                           f' expected from {self.port}')

        ret = self.from_wire(data)
        self._result = ret
        return self

    def __repr__(self) -> str:
        """
        Returns string representation of the command.
        """
        return f'Command: {self._code}, request: {self._data},' \
            f' response: {self.result}'


BaseCommandsT = G90Commands
BaseCommandsDataT = List[Any]


@dataclass
class G90Header:
    """
    Represents JSON structure of the header used in base panel commands.

    :meta private:
    """
    code: Optional[int] = None
    data: Optional[BaseCommandsDataT] = None


class G90BaseCommand(G90Command[BaseCommandsT, BaseCommandsDataT]):
    """
    Class for handling base G90 panel commands.
    """
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._resp = G90Header()

    def encode_data(self, data: Optional[BaseCommandsDataT]) -> str:
        """
        Encodes the command data to JSON string.
        """
        if data is None:
            return '""'
        return json.dumps([self._code, data],
                          # No newlines to be inserted
                          indent=None,
                          # No whitespace around entities
                          separators=(',', ':'))

    def decode_data(self, payload: Optional[str]) -> BaseCommandsDataT:
        """
        Decodes the command data from JSON string.
        """
        # Also, panel may report an error supplying specific reason, e.g.
        # command and its arguments that have failed
        if not payload:
            return []
        if payload.startswith('error'):
            error = payload[5:]
            raise G90CommandError(
                f"Command {self._code.name}"
                f" (code={self._code.value}) failed"
                f" with error: '{error}'")

        resp = None
        try:
            resp = json.loads(payload, strict=False)
        except json.JSONDecodeError as exc:
            raise G90Error(
                f"Unable to parse response as JSON: '{payload}'"
            ) from exc

        if not isinstance(resp, list):
            raise G90Error(
                f"Malformed response, 'list' expected: '{payload}'"
            )

        if resp is not None:
            self._resp = G90Header(*resp)
            _LOGGER.debug('Parsed from wire: %s', self._resp)

            if not self._resp.code:
                raise G90Error(f"Missing code in response: '{payload}'")
            # Check there is data if the response is non-empty
            if not self._resp.data:
                raise G90Error(f"Missing data in response: '{payload}'")

            if self._resp.code != self._code:
                raise G90Error(
                    'Wrong response - received code '
                    f"{self._resp.code}, expected code {self._code}")

        return self._resp.data or []

    def to_wire(self) -> bytes:
        """
        Returns the command in wire format.
        """
        wire = bytes(f'ISTART[{self._code},{self._code},{self._data}]IEND\0',
                     'utf-8')
        _LOGGER.debug('Encoded to wire format %s', wire)
        return wire

    def from_wire(self, data: bytes) -> BaseCommandsDataT:
        """
        Parses the response from the alarm panel.
        """
        _LOGGER.debug('To be decoded from wire format %s', data)
        try:
            decoded_data = data.decode('utf-8')
            if not decoded_data.startswith('ISTART'):
                raise G90Error('Missing start marker in data')
            if not decoded_data.endswith('IEND\0'):
                raise G90Error('Missing end marker in data')
            payload = decoded_data[6:-5]
            _LOGGER.debug("Decoded from wire: string '%s'", payload)

            if not payload:
                return []

            # Panel may report the last command has failed
            if payload == 'fail':
                raise G90CommandFailure(
                    f"Command {self._code.name}"
                    f" (code={self._code.value}) failed"
                )

            return self.decode_data(payload)
        except UnicodeDecodeError as exc:
            raise G90Error(
                'Unable to decode response from UTF-8'
            ) from exc

    @property
    def result(self) -> BaseCommandsDataT:
        """
        The result of the command.
        """
        if self._result is None:
            return []
        return self._result



"""
Tests for network configuration retrieval and modification.
"""
from __future__ import annotations
from typing import Union, Optional
import pytest
from pyg90alarm.alarm import (
    G90Alarm,
)
from pyg90alarm.local.net_config import (
    G90NetConfig, G90APNAuth
)
from .device_mock import DeviceMock


@pytest.mark.parametrize('is_gsm_operator_expected', [
    pytest.param(
        True, marks=pytest.mark.g90device(sent_data=[
            b'ISTART[212,'
            b'[0,"123456789",1,1,"apn.a.net","user","pwd",3,"54321"]'
            b']IEND\0',
            b'ISTARTIEND\0'
        ]),
        id='Device with cellular module'
    ),
    pytest.param(
        False, marks=pytest.mark.g90device(sent_data=[
            b'ISTART[212,'
            b'[0,"123456789",1,1,"apn.a.net","user","pwd",3]'
            b']IEND\0',
            b'ISTARTIEND\0'
        ]),
        id='Device with no cellular module'
    ),
])
async def test_net_config(
    is_gsm_operator_expected: bool,
    mock_device: DeviceMock
) -> None:
    """
    Tests for retrieving and modifying network configuration from
    the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    # Retrieve configuration
    cfg = await g90.net_config()
    assert isinstance(cfg, G90NetConfig)

    # Verify retrieved data
    assert cfg.ap_enabled is False
    assert cfg.ap_password == "123456789"
    assert cfg.wifi_enabled is True
    assert cfg.gprs_enabled is True
    assert cfg.apn_name == "apn.a.net"
    assert cfg.apn_user == "user"
    assert cfg.apn_password == "pwd"
    assert cfg.apn_auth == G90APNAuth.PAP_OR_CHAP
    # Verify presence of optional `gsm_operator` field for devices with
    # cellular module
    if is_gsm_operator_expected:
        assert cfg.gsm_operator == '54321'
    else:
        assert cfg.gsm_operator is None

    # Modify and save the configuration
    cfg.ap_enabled = True
    await cfg.save()

    # Verify data sent to the device, `gsm_operator` field should not be sent
    # regardless if it was present when loading from device
    assert await mock_device.recv_data == [
        b'ISTART[212,212,""]IEND\0',
        b'ISTART[213,213,[213,'
        b'[1,"123456789",1,1,"apn.a.net","user","pwd",3]'
        b']]IEND\0'
    ]


@pytest.mark.parametrize(
    'field_name,invalid_value_low,invalid_value_high,valid_value', [
        pytest.param(
            'apn_name', None, 'a' * 101, 'valid.apn.net',
            id='apn_name'
        ),
        pytest.param(
            'apn_user', None, 'a' * 65, 'valid_user',
            id='apn_user'
        ),
        pytest.param(
            'apn_password', None, 'a' * 65, 'valid_password',
            id='apn_password'
        ),
        pytest.param(
            'ap_enabled', -1, 2, 1,
            id='ap_enabled'
        ),
        pytest.param(
            'wifi_enabled', -1, 2, 0,
            id='wifi_enabled'
        ),
        pytest.param(
            'gprs_enabled', -1, 2, 1,
            id='gprs_enabled'
        ),
        pytest.param(
            'ap_password', 'a' * 8, 'a' * 65, 'valid_password',
            id='ap_password'
        ),
        # Feeding invalid values to enums will not get to the validation, hence
        # underlying integer field is tested directly
        pytest.param(
            '_apn_auth', min(G90APNAuth) - 1, max(G90APNAuth) + 1,
            G90APNAuth.NONE,
            id='apn_auth'
        ),
    ]
)
@pytest.mark.g90device(sent_data=[
    b'ISTART[212,'
    b'[0,"123456789",1,1,"apn.a.net","user","pwd",3,"54321"]'
    b']IEND\0',
    b'ISTARTIEND\0'
])
async def test_net_config_constraints(
    field_name: str,
    invalid_value_low: Optional[Union[int, str]],
    invalid_value_high: Union[int, str],
    valid_value: Union[int, str],
    mock_device: DeviceMock
) -> None:
    """
    Tests for network configuration field validation constraints.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    # Retrieve configuration
    cfg = await g90.net_config()
    assert isinstance(cfg, G90NetConfig)

    # Test setting invalid low value for the fields having minimum length
    # constraint
    if invalid_value_low is not None:
        with pytest.raises(ValueError):
            setattr(cfg, field_name, invalid_value_low)

    # Test setting invalid high value
    with pytest.raises(ValueError):
        setattr(cfg, field_name, invalid_value_high)

    # Test setting valid value
    setattr(cfg, field_name, valid_value)
    assert getattr(cfg, field_name) == valid_value


@pytest.mark.g90device(sent_data=[
    b'ISTART[212,'
    b'[0,"123456789",1,1,"","user","pwd",3,"54321"]'
    b']IEND\0',
    b'ISTARTIEND\0'
])
async def test_net_config_apn_name_empty(
    mock_device: DeviceMock, caplog: pytest.LogCaptureFixture
) -> None:
    """
    Tests for handling network configuration with empty APN name.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    # Retrieve configuration
    cfg = await g90.net_config()
    assert isinstance(cfg, G90NetConfig)

    # Verify retrieved values
    assert cfg.apn_name == ''

    # Ensure that no validation error was logged for empty APN name, the value
    # of the field as received from the panel is trusted, so no exception will
    # be raised only error logged
    assert (
        'apn_name: Validation failed during initialization for trusted value'
    ) not in ''.join(caplog.messages)

    # Verify empty APN name is allowed to be sent back to the panel
    await cfg.save()

    assert await mock_device.recv_data == [
        b'ISTART[212,212,""]IEND\0',
        b'ISTART[213,213,[213,'
        b'[0,"123456789",1,1,"","user","pwd",3]'
        b']]IEND\0'
    ]


@pytest.mark.parametrize('new_apn_auth_value,expected_recv_data', [
    # Verify the invalid value is retained, since the panel is trusted, if no
    # modifications to the value is made
    pytest.param(
        None, [
            b'ISTART[212,212,""]IEND\0',
            b'ISTART[213,213,[213,'
            b'[0,"123456789",1,1,"","user","pwd",333]'
            b']]IEND\0'
        ], id='Device with invalid APN auth value unmodified'
    ),
    # Verify the invalid value is replaced with valid one when modified
    pytest.param(
        G90APNAuth.PAP, [
            b'ISTART[212,212,""]IEND\0',
            b'ISTART[213,213,[213,'
            b'[0,"123456789",1,1,"","user","pwd",1]'
            b']]IEND\0'
        ], id='Device with invalid APN auth value modified to valid value'
    ),
])
@pytest.mark.g90device(sent_data=[
    b'ISTART[212,'
    b'[0,"123456789",1,1,"","user","pwd",333,"54321"]'
    b']IEND\0',
    b'ISTARTIEND\0'
])
async def test_net_config_apn_auth_invalid(
    new_apn_auth_value: Optional[G90APNAuth],
    expected_recv_data: list[bytes],
    mock_device: DeviceMock
) -> None:
    """
    Tests for handling network configuration with invalid APN authentication
    method value.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    # Retrieve configuration
    cfg = await g90.net_config()
    assert isinstance(cfg, G90NetConfig)

    # Verify retrieved value is mapped to NONE enum value
    assert cfg.apn_auth == G90APNAuth.NONE

    # Modify the value if requested
    if new_apn_auth_value is not None:
        cfg.apn_auth = new_apn_auth_value
    # Save the configuration
    await cfg.save()

    # Verify data sent to the device
    assert await mock_device.recv_data == expected_recv_data



# Copyright (c) 2026 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Provides support for system commands of the G90 alarm panel.
"""
from __future__ import annotations
from typing import Any, Optional, TypeVar, List
import logging

from ..const import G90SystemCommands, G90SystemConfigurationCommands
from .base_cmd import G90Command

_LOGGER = logging.getLogger(__name__)


SystemCommandsT = TypeVar('SystemCommandsT', bound=G90SystemCommands)
SystemCommandsDataT = TypeVar('SystemCommandsDataT')


class G90SystemCommandBase(G90Command[SystemCommandsT, SystemCommandsDataT]):
    """
    Base class for system commands of the G90 alarm panel.

    :param code: System command code
    :param kwargs: Additional arguments passed to the base class
    """
    def __init__(
        self, *, code: SystemCommandsT, **kwargs: Any
    ) -> None:
        super().__init__(code=code, **kwargs)

    @property
    def expects_response(self) -> bool:
        """
        Indicates whether the command expects a response.

        System commands do not return any response.

        :return: False
        """
        return False

    def encode_data(self, data: Optional[SystemCommandsDataT]) -> str:
        """
        Encodes the command data to string.

        :param data: Data for the command
        :return: Encoded data string
        """
        # Placeholder as required by the base class
        raise NotImplementedError()  # pragma: no cover

    def decode_data(self, payload: Optional[str]) -> SystemCommandsDataT:
        """
        Decodes the command data from string.

        :param payload: Payload string
        :return: Decoded data
        """
        # Placeholder as required by the base class
        raise NotImplementedError()  # pragma: no cover

    def to_wire(self) -> bytes:
        """
        Serializes the command to wire format as expected by the panel.

        :return: Wire data
        """
        wire = bytes(
            f'ISTART[0,100,"AT^IWT={self._code}{self._data},IWT"]IEND\0',
            'utf-8'
        )

        _LOGGER.debug('Encoded to wire format %s', wire)
        return wire

    def from_wire(self, data: bytes) -> SystemCommandsDataT:
        """
        Deserializes the command from wire format.

        :param data: Wire data
        :return: Decoded data
        """
        # Placeholder as required by the base class
        raise NotImplementedError()  # pragma: no cover

    @property
    def result(self) -> SystemCommandsDataT:
        """
        The result of the command.

        :return: Decoded data
        """
        # Placeholder as required by the base class
        raise NotImplementedError()  # pragma: no cover


class G90SystemCommand(
    G90SystemCommandBase[SystemCommandsT, str]
):
    """
    Represents a system command of the G90 alarm panel.

    :param code: System command code
    :param kwargs: Additional arguments passed to the base class
    """
    def __init__(
        self, *, code: SystemCommandsT, **kwargs: Any
    ) -> None:
        if code in [G90SystemCommands.SET_CONFIGURATION,
                    G90SystemCommands.GET_CONFIGURATION]:
            raise ValueError(
                'Use G90SystemConfigurationCommand class for '
                'SET_CONFIGURATION and GET_CONFIGURATION commands'
            )
        super().__init__(code=code, **kwargs)

    def encode_data(self, data: Optional[str]) -> str:
        """
        Encodes the command data to string.

        :param data: Data for the command
        :return: Encoded data string
        """
        return data or ''

    def decode_data(self, payload: Optional[str]) -> str:
        """
        Decodes the command data from string.

        No response is expected for system commands.

        :param payload: Payload string (ignored)
        :return: Empty string
        """
        # No response is expected for system commands
        return ''  # pragma: no cover

    def from_wire(self, data: bytes) -> str:
        """
        Deserializes the command from wire format.

        No response is expected for system commands.

        :param data: Wire data (ignored)
        :return: Empty string
        """
        # No response is expected for system commands
        return ''  # pragma: no cover

    @property
    def result(self) -> str:
        """
        The result of the command.

        No response is expected for system commands.

        :return: Empty string
        """
        # No response is expected for system commands
        return ''


class G90SystemConfigurationCommand(
    G90SystemCommandBase[G90SystemCommands, List[str]]
):
    """
    Represents a system configuration command of the G90 alarm panel.

    :param cmd: Sub-command code for configuration command
    :param data: Data for the command
    :param kwargs: Additional arguments passed to the base class
    """
    def __init__(
        self, *, cmd: G90SystemConfigurationCommands,
        data: Optional[List[str]] = None,
        **kwargs: Any
    ) -> None:
        self._cmd = cmd
        super().__init__(
            code=G90SystemCommands.SET_CONFIGURATION, data=data, **kwargs
        )

    def encode_data(self, data: Optional[List[str]]) -> str:
        """
        Encodes the command data to string.

        :param data: Data for the command
        :return: Encoded data string
        """
        if data is None:
            raise ValueError('Data must be provided for configuration command')

        res = f',{self._cmd.value}={"&".join(data)}'
        _LOGGER.debug('Encoded data "%s" to string format: %s', data, res)
        return res

    def decode_data(self, payload: Optional[str]) -> List[str]:
        """
        Decodes the command data from string.

        System configuration commands do not return any response.

        :param payload: Payload string
        :return: Empty list
        """
        # System configuration commands do not return any response.
        return []  # pragma: no cover

    def from_wire(self, data: bytes) -> List[str]:
        """
        Deserializes the command from wire format.

        System configuration commands do not return any response.

        :param data: Wire data
        :return: Empty list
        """
        # System configuration commands do not return any response.
        return []  # pragma: no cover

    @property
    def result(self) -> List[str]:
        """
        The result of the command.

        System configuration commands do not return any response.

        :return: Empty list
        """
        # System configuration commands do not return any response.
        return []  # pragma: no cover


class G90SetServerAddressCommand(
    G90SystemConfigurationCommand
):
    """
    Sets the server address for the panel to communicate with using the cloud
    notifications protocol.

    :param cloud_ip: IP address of the cloud server
    :param cloud_port: Port number of the cloud server
    :param kwargs: Additional arguments passed to the base class
    """
    def __init__(
        self, *, cloud_ip: str, cloud_port: int,
        **kwargs: Any
    ) -> None:
        super().__init__(
            cmd=G90SystemConfigurationCommands.SERVER_ADDRESS,
            # The command requires two fields for the cloud server address, 1st
            # is used by the panel to communicate with the server, while 2nd
            # might potentially be a fallback address or alike. However,
            # experiments did not show any attempts by the panel to use the 2nd
            # address, so we set both to the same value in an avoidance of
            # doubt.
            data=[cloud_ip, cloud_ip, str(cloud_port)], **kwargs
        )



"""
Tests for G90BaseList class
"""
from __future__ import annotations
from typing import AsyncGenerator
from unittest.mock import MagicMock

from pyg90alarm.entities.base_list import G90BaseList


async def test_find_free_idx_empty_list() -> None:
    """
    Tests find_free_idx with empty entities list.
    """
    class TestClass(G90BaseList[MagicMock]):
        """
        Mock subclass for testing G90BaseList.
        """
        async def _fetch(self) -> AsyncGenerator[MagicMock, None]:
            """
            Mock no entities.
            """
            x: MagicMock
            for x in []:
                yield x

    base_list = TestClass(parent=MagicMock())
    result = await base_list.find_free_idx()

    assert result == 0


async def test_find_free_idx_one_entity_at_zero() -> None:
    """
    Tests find_free_idx with one entity at index 0.
    """
    class TestClass(G90BaseList[MagicMock]):
        """
        Mock subclass for testing G90BaseList.
        """
        async def _fetch(self) -> AsyncGenerator[MagicMock, None]:
            """
            Mock one entity at index 0.
            """
            for x in [MagicMock(index=0)]:
                yield x

    base_list = TestClass(parent=MagicMock())
    result = await base_list.find_free_idx()

    assert result == 1


async def test_find_free_idx_returns_lowest_available_index() -> None:
    """
    Tests find_free_idx with two entities at indexes 10 and 11.
    """
    class TestClass(G90BaseList[MagicMock]):
        """
        Mock subclass for testing G90BaseList.
        """
        async def _fetch(self) -> AsyncGenerator[MagicMock, None]:
            """
            Mock entities at indexes 10 and 11.
            """
            for x in [MagicMock(index=10), MagicMock(index=11)]:
                yield x

    base_list = TestClass(parent=MagicMock())
    result = await base_list.find_free_idx()

    assert result == 0


async def test_find_free_idx_returns_lowest_available_index_over_gap() -> None:
    """
    Tests find_free_idx with two entities at indexes 10 and 11.
    """
    class TestClass(G90BaseList[MagicMock]):
        """
        Mock subclass for testing G90BaseList.
        """
        async def _fetch(self) -> AsyncGenerator[MagicMock, None]:
            """
            Mock entities at indexes 0, 10 and 11.
            """
            for x in [
                MagicMock(index=0), MagicMock(index=10), MagicMock(index=11)
            ]:
                yield x

    base_list = TestClass(parent=MagicMock())
    result = await base_list.find_free_idx()

    assert result == 1


async def test_find_entity_by_idx_and_name() -> None:
    """
    Tests find with matching index, subindex, and name.
    """
    entity = MagicMock()
    entity.index = 0
    entity.subindex = 0
    entity.name = "Test Entity"
    entity.is_unavailable = False

    class TestClass(G90BaseList[MagicMock]):
        """
        Mock subclass for testing G90BaseList.
        """
        async def _fetch(self) -> AsyncGenerator[MagicMock, None]:
            """
            Mock one entity.
            """
            yield entity

    base_list = TestClass(parent=MagicMock())
    result = await base_list.find(
        idx=0, name="Test Entity", exclude_unavailable=False
    )

    assert result == entity


async def test_find_entity_name_mismatch() -> None:
    """
    Tests find with matching index but mismatched name.
    """
    class TestClass(G90BaseList[MagicMock]):
        """
        Mock subclass for testing G90BaseList.
        """
        async def _fetch(self) -> AsyncGenerator[MagicMock, None]:
            """
            Mock one entity.
            """
            entity = MagicMock()
            entity.index = 0
            entity.subindex = 0
            entity.name = "Test Entity"
            entity.is_unavailable = False
            yield entity

    base_list = TestClass(parent=MagicMock())
    result = await base_list.find(
        idx=0, name="Wrong Name", exclude_unavailable=False
    )

    assert result is None


async def test_find_entity_not_found() -> None:
    """
    Tests find with non-existing index.
    """
    class TestClass(G90BaseList[MagicMock]):
        """
        Mock subclass for testing G90BaseList.
        """
        async def _fetch(self) -> AsyncGenerator[MagicMock, None]:
            """
            Mock no entities.
            """
            x: MagicMock
            for x in []:
                yield x

    base_list = TestClass(parent=MagicMock())
    result = await base_list.find(
        idx=0, name="Test Entity", exclude_unavailable=False
    )

    assert result is None


async def test_find_entity_unavailable_excluded() -> None:
    """
    Tests find with unavailable entity and exclude_unavailable=True.
    """
    class TestClass(G90BaseList[MagicMock]):
        """
        Mock subclass for testing G90BaseList.
        """
        async def _fetch(self) -> AsyncGenerator[MagicMock, None]:
            """
            Mock one entity.
            """
            entity = MagicMock()
            entity.index = 0
            entity.subindex = 0
            entity.name = "Test Entity"
            entity.is_unavailable = True
            yield entity

    base_list = TestClass(parent=MagicMock())
    result = await base_list.find(
        idx=0, name="Test Entity", exclude_unavailable=True
    )

    assert result is None


async def test_find_entity_unavailable_not_excluded() -> None:
    """
    Tests find with unavailable entity and exclude_unavailable=False.
    """
    entity = MagicMock()
    entity.index = 0
    entity.subindex = 0
    entity.name = "Test Entity"
    entity.is_unavailable = True

    class TestClass(G90BaseList[MagicMock]):
        """
        Mock subclass for testing G90BaseList.
        """
        async def _fetch(self) -> AsyncGenerator[MagicMock, None]:
            """
            Mock one entity.
            """
            yield entity

    base_list = TestClass(parent=MagicMock())
    result = await base_list.find(
        idx=0, name="Test Entity", exclude_unavailable=False
    )

    assert result == entity


async def test_find_entity_with_subindex() -> None:
    """
    Tests find with specific subindex.
    """
    entity = MagicMock()
    entity.index = 0
    entity.subindex = 1
    entity.name = "Test Entity"
    entity.is_unavailable = False

    class TestClass(G90BaseList[MagicMock]):
        """
        Mock subclass for testing G90BaseList.
        """
        async def _fetch(self) -> AsyncGenerator[MagicMock, None]:
            """
            Mock one entity.
            """
            yield entity

    base_list = TestClass(parent=MagicMock())
    result = await base_list.find(
        idx=0, name="Test Entity", exclude_unavailable=False, subindex=1
    )

    assert result == entity


async def test_duplicate_entities() -> None:
    """
    Tests that duplicate entities are handled correctly during update.
    """
    existing_entity = MagicMock()
    existing_entity.index = 1
    existing_entity.subindex = 0
    existing_entity.name = "Existing entity"
    existing_entity.is_unavailable = False

    new_entity = MagicMock()
    new_entity.index = 1
    new_entity.subindex = 0
    new_entity.name = "Test entity"
    new_entity.is_unavailable = False

    mock_entities = iter([
        # Initial fetch returns one existing entity
        [existing_entity],
        # Second fetch returns duplicate new entities
        [new_entity, new_entity],
    ])

    class TestClass(G90BaseList[MagicMock]):
        """
        Mock subclass for testing G90BaseList.
        """
        async def _fetch(self) -> AsyncGenerator[MagicMock, None]:
            """
            Mock test list entities.
            """
            # Yield entities under test
            for x in next(mock_entities):
                yield x

    base_list = TestClass(parent=MagicMock())
    # Initial update to simulate the existing entity
    await base_list.update()
    # Second update to test handling of duplicates, should not raise
    result = await base_list.update()

    assert len(result) == 2
    # Initial entities not present in second update result should be marked
    # unavailable
    assert result[0] == existing_entity
    assert result[0].is_unavailable is True
    assert result[1] == new_entity



"""
Tests for package-specific exceptions.
"""
from __future__ import annotations
import pytest

from pyg90alarm.exceptions import G90TimeoutError, G90Error


def test_g90_timeout_error_inheritance() -> None:
    """
    Test that catching G90Error also catches G90TimeoutError.
    """

    with pytest.raises(G90Error):
        raise G90TimeoutError("Timeout occurred")



# Copyright (c) 2026 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Unit tests for ReadOnlyIfNotProvided descriptor, field_readonly_if_not_provided
function.
"""
from __future__ import annotations
from dataclasses import dataclass, fields
from typing import Any, Optional

import pytest

from pyg90alarm.dataclass.load_save import (
    DataclassLoadSave,
    Metadata,
    field_readonly_if_not_provided,
)


@dataclass
class SimpleReadOnlyConfig(DataclassLoadSave):
    """
    Test dataclass with read-only field.
    """
    read_only_field: Optional[int] = field_readonly_if_not_provided(
        default=None
    )
    regular_field: int = 0


@pytest.mark.parametrize(
    "default_value,expected_default",
    [
        pytest.param(None, None, id="None default"),
        pytest.param(42, 42, id="int default"),
        pytest.param("test", "test", id="str default"),
        pytest.param([1, 2, 3], [1, 2, 3], id="list default"),
        pytest.param({"key": "value"}, {"key": "value"}, id="dict default"),
    ],
)
def test_readonly_descriptor_read_not_provided(
    default_value: Any, expected_default: Any
) -> None:
    """
    Test reading a read-only field that was not provided during initialization.
    """

    @dataclass
    class TestClass:
        field_ro: Any = field_readonly_if_not_provided(default=default_value)

    obj = TestClass()
    assert obj.field_ro == expected_default


@pytest.mark.parametrize(
    "init_value,expected_value",
    [
        pytest.param(42, 42, id="int value"),
        pytest.param("hello", "hello", id="str value"),
        pytest.param([1, 2], [1, 2], id="list value"),
        pytest.param(0, 0, id="zero value"),
        pytest.param("", "", id="empty string"),
    ],
)
def test_readonly_descriptor_read_provided(
    init_value: Any, expected_value: Any
) -> None:
    """
    Test reading a read-only field that was provided during initialization.
    """

    @dataclass
    class TestClass:
        field_ro: Any = field_readonly_if_not_provided(default=None)

    obj = TestClass(field_ro=init_value)
    assert obj.field_ro == expected_value


def test_readonly_descriptor_write_not_provided_raises_error() -> None:
    """
    Test that writing to a non-provided read-only field raises ValueError.
    """
    obj = SimpleReadOnlyConfig()

    with pytest.raises(ValueError):
        obj.read_only_field = 100


def test_readonly_descriptor_write_provided_succeeds() -> None:
    """
    Test that writing to a provided read-only field succeeds.
    """
    obj = SimpleReadOnlyConfig(read_only_field=42)

    # Should not raise
    obj.read_only_field = 100
    assert obj.read_only_field == 100


@pytest.mark.parametrize(
    "new_value",
    [
        pytest.param(123, id="int value"),
        pytest.param(0, id="zero"),
        pytest.param(-1, id="negative"),
        pytest.param(None, id="None"),
    ],
)
def test_readonly_descriptor_write_provided_multiple_values(
    new_value: Any,
) -> None:
    """
    Test writing multiple values to a provided read-only field.
    """
    obj = SimpleReadOnlyConfig(read_only_field=1)

    obj.read_only_field = new_value
    assert obj.read_only_field == new_value


def test_readonly_descriptor_multiple_instances_independent() -> None:
    """
    Test that read-only descriptors work independently for different instances.
    """
    obj1 = SimpleReadOnlyConfig(read_only_field=10)
    obj2 = SimpleReadOnlyConfig()

    # obj1 should allow writes
    obj1.read_only_field = 20
    assert obj1.read_only_field == 20

    # obj2 should raise on write
    with pytest.raises(ValueError):
        obj2.read_only_field = 30

    # Values should be independent
    assert obj1.read_only_field == 20
    assert obj2.read_only_field is None


def test_readonly_descriptor_error_message_includes_field_name() -> None:
    """
    Test that ValueError message includes the correct field name.
    """
    obj = SimpleReadOnlyConfig()

    with pytest.raises(ValueError) as exc_info:
        obj.read_only_field = 100

    assert "read_only_field" in str(exc_info.value)
    assert "read-only" in str(exc_info.value)


def test_field_readonly_creates_dataclass_field() -> None:
    """
    Test that field_readonly_if_not_provided returns a valid dataclass field.
    """

    @dataclass
    class TestClass:
        field_ro: int = field_readonly_if_not_provided(default=42)

    obj = TestClass()
    assert obj.field_ro == 42


def test_field_readonly_with_none_default_sets_skip_none_metadata() -> None:
    """
    Test that SKIP_NONE metadata is set when default is None.
    """

    @dataclass
    class TestClass(DataclassLoadSave):
        field_ro: Optional[int] = field_readonly_if_not_provided(default=None)

    field_meta = fields(TestClass)[0]
    assert Metadata.SKIP_NONE in field_meta.metadata
    assert field_meta.metadata[Metadata.SKIP_NONE] is True


def test_field_readonly_with_non_none_default_no_skip_none() -> None:
    """
    Test that SKIP_NONE is not set when default is not None.
    """

    @dataclass
    class TestClass:
        field_ro: int = field_readonly_if_not_provided(default=42)

    field_meta = fields(TestClass)[0]
    assert Metadata.SKIP_NONE not in field_meta.metadata


def test_field_readonly_preserves_existing_metadata() -> None:
    """
    Test that field_readonly_if_not_provided preserves existing metadata.
    """

    @dataclass
    class TestClass:
        field_ro: Optional[int] = field_readonly_if_not_provided(
            default=None, metadata={Metadata.NO_SERIALIZE: True}
        )

    field_meta = fields(TestClass)[0]
    assert Metadata.SKIP_NONE in field_meta.metadata
    assert Metadata.NO_SERIALIZE in field_meta.metadata


def test_field_readonly_with_custom_metadata() -> None:
    """
    Test field_readonly_if_not_provided with custom metadata dict.
    """

    @dataclass
    class TestClass:
        field_ro: Optional[str] = field_readonly_if_not_provided(
            default=None, metadata={"custom_key": "custom_value"}
        )

    field_meta = fields(TestClass)[0]
    assert field_meta.metadata["custom_key"] == "custom_value"
    assert Metadata.SKIP_NONE in field_meta.metadata


def test_multiple_readonly_fields_independent() -> None:
    """
    Test that multiple read-only fields are independent.
    """

    @dataclass
    class TestClass(DataclassLoadSave):
        """
        Test dataclass with multiple read-only fields.
        """
        required_field: int = field_readonly_if_not_provided(
            default=42
        )
        optional_field: Optional[str] = field_readonly_if_not_provided(
            default=None
        )
        regular_field: str = "test"

    obj = TestClass(required_field=10, optional_field="test")

    # Both should be writable
    obj.required_field = 20
    obj.optional_field = "changed"

    assert obj.required_field == 20
    assert obj.optional_field == "changed"




# Copyright (c) 2026 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Base class for loading/saving dataclasses to a device.
"""
from __future__ import annotations
from typing import (
    TYPE_CHECKING, Type, TypeVar, Optional, ClassVar, Any, Dict, List, cast
)
import logging
from dataclasses import dataclass, asdict, field, fields
from .validation import ValidatorBase
from ..const import G90Commands
if TYPE_CHECKING:
    from ..alarm import G90Alarm


_LOGGER = logging.getLogger(__name__)
DataclassLoadSaveT = TypeVar('DataclassLoadSaveT', bound='DataclassLoadSave')
T = TypeVar('T')


class Metadata:
    """
    Metadata keys for DataclassLoadSave fields.
    """
    # pylint: disable=too-few-public-methods
    NO_SERIALIZE = 'no_serialize'
    SKIP_NONE = 'skip_none'


class ReadOnlyIfNotProvided(ValidatorBase[T]):
    """
    Descriptor for dataclass fields to be read-only if not provided during
    initialization.

    The field can be read, but attempts to modify it will raise a
    ValueError if the field was not provided during initialization. In
    other words, the only way to set the value is during object creation.

    Example usage:

        @dataclass
        class Example:
            read_only_field: Optional[int] = field_readonly_if_not_provided(
                default=None
            )

        # Works ok
        ex = Example(read_only_field=42)
        print(ex.read_only_field)  # Outputs: 42
        ex.read_only_field = 100  # Works ok

        # Raises ValueError
        ex2 = Example()
        print(ex2.read_only_field)  # Outputs: None
        ex2.read_only_field = 100  # Raises ValueError

    :param default: Default value to return upon read if not provided during
     initialization.
    """
    # pylint: disable=too-few-public-methods

    def __validate__(self, obj: Any, value: T) -> bool:
        """
        Validation method.
        """
        # Prevent setting the value if it was not provided during
        # initialization. The condition is determined by checking if the
        # current value is `self` - i.e. the descriptor instance hasn't been
        # replaced with an actual value
        if getattr(obj, self.__field_name__, self._default) is self:
            raise ValueError(
                f'Field {self.__unmangled_name__} is read-only because'
                ' it was not provided during initialization'
            )
        return True


def field_readonly_if_not_provided(
    *args: Any, default: Optional[T] = None, **kwargs: Any
) -> T:
    """
    Helper function to create a dataclass field with ReadOnlyIfNotProvided
    descriptor.

    :param args: Positional arguments to pass to `dataclasses.field()`.
    :param default: Default value to return upon read if not provided during
     initialization.
    :param kwargs: Keyword arguments to pass to `dataclasses.field()`.
    :return: A dataclass field with ReadOnlyIfNotProvided descriptor attached.
    """
    # Also set SKIP_NONE metadata if default is None, so that the field is
    # skipped during serialization when its value is None
    if default is None:
        if 'metadata' not in kwargs:
            kwargs['metadata'] = {}
        kwargs['metadata'][Metadata.SKIP_NONE] = True

    # Instantiate the field with ReadOnlyIfNotProvided descriptor and rest of
    # the provided arguments
    # pylint: disable=invalid-field-call
    return cast(T, field(
        *args, **kwargs,
        default=ReadOnlyIfNotProvided[T](default)
    ))


@dataclass
class DataclassLoadSave:
    """
    Base class for loading/saving dataclasses to a device.

    There are multiple ways to implement the functionality:
     - Encapsulate the dataclass inside another class that handles
       loading/saving and exposes dataclass fields as properties. The latter
       part gets complex as properties need to be asynchronous, as well as
       added dynamically at runtime to improve maintainability.
     - Inherit from this class, which provides `load` and `save` methods on top
       of standard dataclasses. This is believed to be more concise and easier
       to understand.

    Implementing classes must define `LOAD_COMMAND` and `SAVE_COMMAND` class
    variables to specify which commands to use for loading and saving data.

    Example usage:

        @dataclass
        class G90ExampleConfig(DataclassLoadSave):
            LOAD_COMMAND = G90Commands.GETEXAMPLECONFIG
            SAVE_COMMAND = G90Commands.SETEXAMPLECONFIG
            field1: int
            field2: str

        # Loading data
        config = await G90ExampleConfig.load(G90_alarm_instance)
        print(config.field1, config.field2)

        # Modifying and saving data
        config.field1 = 42
        await config.save()
    """
    LOAD_COMMAND: ClassVar[Optional[G90Commands]] = None
    SAVE_COMMAND: ClassVar[Optional[G90Commands]] = None

    def __post_init__(self) -> None:
        """
        Post-initialization processing.
        """
        # Instance variable to hold reference to parent G90Alarm instance,
        # declared here to avoid being part of dataclass fields
        self._parent: Optional[G90Alarm] = None

    def serialize(self) -> List[Any]:
        """
        Returns the dataclass fields as a list.

        Handles specific metadata for the fields.
        :seealso:`Metadata`.

        :return: Dataclass serialized as list.
        """
        result = []

        for f in fields(self):
            # Skip fields marked with NO_SERIALIZE metadata
            if f.metadata.get(Metadata.NO_SERIALIZE, False):
                continue

            # Skip fields with None value if SKIP_NONE metadata is set
            if (
                f.metadata.get(Metadata.SKIP_NONE, False)
                and getattr(self, f.name) is None
            ):
                continue

            # Append field value to the result list
            result.append(getattr(self, f.name))

        return result

    async def save(self) -> None:
        """
        Save the current data to the device.
        """
        assert self.SAVE_COMMAND is not None, '`SAVE_COMMAND` must be defined'
        assert self._parent is not None, 'Please call `load()` first'

        _LOGGER.debug('Setting data to the device: %s', str(self))
        await self._parent.command(
            self.SAVE_COMMAND,
            self.serialize()
        )

    @classmethod
    async def load(
        cls: Type[DataclassLoadSaveT], parent: G90Alarm
    ) -> DataclassLoadSaveT:
        """
        Create an instance with values loaded from the device.

        :return: An instance of the dataclass loaded from the device.
        """
        assert cls.LOAD_COMMAND is not None, '`LOAD_COMMAND` must be defined'
        assert parent is not None, '`parent` must be provided'

        data = await parent.command(cls.LOAD_COMMAND)
        obj = cls(*data)
        _LOGGER.debug('Loaded data: %s', str(obj))

        obj._parent = parent

        return obj

    def _asdict(self) -> Dict[str, Any]:
        """
        Returns the dataclass fields as a dictionary.

        :return: A dictionary representation.
        """
        return asdict(self)

    def __str__(self) -> str:
        """
        Textual representation of the entry.

        `str()` is used instead of `repr()` since dataclass provides `repr()`
        by default, and it would be impractical to require each ancestor to
        disable that.

        :return: A textual representation.
        """
        return super().__repr__() + f'({str(self._asdict())})'



"""
Tests for alarm panel configuration retrieval and modification.
"""
from __future__ import annotations
import pytest
from pyg90alarm.alarm import (
    G90Alarm,
)
from pyg90alarm.local.host_config import (
    G90HostConfig, G90SpeechLanguage, G90VolumeLevel,
)
from .device_mock import DeviceMock


@pytest.mark.parametrize('is_ring_volume_expected,expected_recv_data', [
    pytest.param(
        True, [
            b'ISTART[106,106,""]IEND\0',
            b'ISTART[107,107,[107,'
            b'[600,0,0,1,2,2,60,2,0,60,2]'
            b']]IEND\0'
        ], marks=pytest.mark.g90device(sent_data=[
            b'ISTART[106,'
            b'[900,0,0,1,2,2,60,2,0,60,2]'
            b']IEND\0',
            b'ISTARTIEND\0'
        ]),
        id='Device with cellular module'
    ),
    pytest.param(
        False, [
            b'ISTART[106,106,""]IEND\0',
            b'ISTART[107,107,[107,'
            b'[600,0,0,1,2,2,60,2,0,60]'
            b']]IEND\0'
        ], marks=pytest.mark.g90device(sent_data=[
            b'ISTART[106,'
            b'[900,0,0,1,2,2,60,2,0,60]'
            b']IEND\0',
            b'ISTARTIEND\0'
        ]),
        id='Device with no cellular module'
    ),
])
async def test_host_config(
    is_ring_volume_expected: bool, expected_recv_data: list[bytes],
    mock_device: DeviceMock
) -> None:
    """
    Tests for retrieving and modifying alarm panel configuration from
    the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    # Retrieve configuration
    cfg = await g90.host_config()
    assert isinstance(cfg, G90HostConfig)

    # Verify retrieved values
    assert cfg.alarm_siren_duration == 900
    assert cfg.speech_language == G90SpeechLanguage.ENGLISH_MALE
    assert cfg.speech_volume_level == G90VolumeLevel.HIGH
    assert cfg.timezone_offset_m == 60
    assert cfg.key_tone_volume_level == G90VolumeLevel.MUTE
    assert cfg.ring_duration == 60
    assert cfg.alarm_delay == 0
    assert cfg.arm_delay == 0
    assert cfg.backlight_duration == 1
    assert cfg.alarm_volume_level == G90VolumeLevel.HIGH
    # The optional `ring_volume_level` field should be present only if the
    # device has cellular module
    if is_ring_volume_expected:
        assert cfg.ring_volume_level == G90VolumeLevel.HIGH
    else:
        assert cfg.ring_volume_level is None

    # Modify and save configuration
    cfg.alarm_siren_duration = 600
    await cfg.save()

    # Verify data sent to the device
    assert await mock_device.recv_data == expected_recv_data


@pytest.mark.parametrize(
    'field_name,invalid_value_low,invalid_value_high,valid_value', [
        pytest.param(
            'alarm_siren_duration', -1, 2551, 500,
            id='alarm_siren_duration'
        ),
        pytest.param(
            'ring_duration', -1, 2551, 200,
            id='ring_duration'
        ),
        pytest.param(
            'alarm_delay', -1, 2551, 100,
            id='alarm_delay'
        ),
        pytest.param(
            'arm_delay', -1, 2551, 200,
            id='arm_delay'
        ),
        pytest.param(
            'backlight_duration', -1, 2551, 5,
            id='backlight_duration'
        ),
        pytest.param(
            'timezone_offset_m', -720 - 1, 720 + 1, 0,
            id='timezone_offset_m'
        ),
        # Feeding invalid values to enums will not get to the validation, hence
        # underlying integer fields are tested directly
        pytest.param(
            '_speech_language', min(G90SpeechLanguage) - 1,
            max(G90SpeechLanguage) + 1, G90SpeechLanguage.ENGLISH_FEMALE,
            id='speech_language'
        ),
        pytest.param(
            '_speech_volume_level', min(G90VolumeLevel) - 1,
            max(G90VolumeLevel) + 1, G90VolumeLevel.LOW,
            id='speech_volume_level'
        ),
        pytest.param(
            '_alarm_volume_level', min(G90VolumeLevel) - 1,
            max(G90VolumeLevel) + 1, G90VolumeLevel.MUTE,
            id='alarm_volume_level'
        ),
        pytest.param(
            '_key_tone_volume_level', min(G90VolumeLevel) - 1,
            max(G90VolumeLevel) + 1, G90VolumeLevel.HIGH,
            id='key_tone_volume_level'
        ),
    ]
)
@pytest.mark.g90device(sent_data=[
    b'ISTART[106,'
    b'[900,0,0,1,2,2,60,2,0,60,2]'
    b']IEND\0',
    b'ISTARTIEND\0'
])
async def test_host_config_constraints(
    field_name: str,
    invalid_value_low: int,
    invalid_value_high: int,
    valid_value: int,
    mock_device: DeviceMock
) -> None:
    """
    Tests for validation constraints of host configuration fields.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    # Retrieve configuration
    cfg = await g90.host_config()

    # Test setting invalid low value
    with pytest.raises(ValueError):
        setattr(cfg, field_name, invalid_value_low)

    # Test setting invalid high value
    with pytest.raises(ValueError):
        setattr(cfg, field_name, invalid_value_high)

    # Test setting valid value
    setattr(cfg, field_name, valid_value)
    assert getattr(cfg, field_name) == valid_value


@pytest.mark.g90device(sent_data=[
    b'ISTART[106,'
    b'[900,0,0,1,2,2,60,0,0,60,2]'
    b']IEND\0',
    b'ISTARTIEND\0'
])
async def test_host_config_speech_language_zero(
    mock_device: DeviceMock
) -> None:
    """
    Tests for handling host configuration with speech language set to zero.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    # Retrieve configuration
    cfg = await g90.host_config()
    assert isinstance(cfg, G90HostConfig)

    # Verify retrieved values
    assert cfg.speech_language == G90SpeechLanguage.NONE

    # Verify zero value is allowed to be sent back to the panel
    await cfg.save()

    assert await mock_device.recv_data == [
        b'ISTART[106,106,""]IEND\0',
        b'ISTART[107,107,[107,'
        b'[900,0,0,1,2,2,60,0,0,60,2]'
        b']]IEND\0'
    ]
